/*********************************************************************
*               (c) SEGGER Microcontroller GmbH & Co. KG             *
*                        The Embedded Experts                        *
*                           www.segger.com                           *
**********************************************************************
----------------------------------------------------------------------
File    : FlashPrg.c
Purpose : Implementation of RAMCode for ST STM32F205RC
--------  END-OF-HEADER  ---------------------------------------------
*/
#include "FlashOS.H"

#define FLASH           ((FLASH_REGS*)(0x40023C00))
//
// Flash Keys
//
#define FLASH_KEY1      0x45670123
#define FLASH_KEY2      0xCDEF89AB

/*********************************************************************
*
*       Types
*
**********************************************************************
*/

typedef struct {
  U32 acr;
} RESTORE_INFO;

typedef union {
  struct {
    void* pSrc;
    void* pDest;
    U32   NumHWords;
    U32   WritePara0;
    U32   WritePara1;
    U32   WritePara2;
  } In;
  struct {
    void* FailAddr;
  } Out;
} IO_PARA;

//
// Flash registers
//
typedef struct {
  volatile U32 ACR;
  volatile U32 KEYR;
  volatile U32 OPTKEYR;
  volatile U32 SR;
  volatile U32 CR;
  volatile U32 OPTCR;
} FLASH_REGS;


/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
//
// We use this dummy variable to make sure that the PrgData
// section is present in the output elf-file as this section
// is mandatory in current versions of the J-Link DLL 
//
static volatile int _Dummy;
static RESTORE_INFO _pRestoreInfo;

/*********************************************************************
*
*       Static code
*
**********************************************************************
*/
/*********************************************************************
*
*       _FeedWatchdog
*
*  Function description
*    Feeds the watchdog. Needs to be called during RAMCode execution
*    in case of an watchdog is active.
*/
static void _FeedWatchdog(void) {
  // TBD
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/*********************************************************************
*
*       Init
*
*  Function description
*    Handles the initialization of the flash module.
*
*  Parameters
*    Addr: Flash base address
*    Freq: Clock frequency in Hz
*    Func: Caller type (e.g.: 1 - Erase, 2 - Program, 3 - Verify)
*
*  Return value 
*    0 O.K.
*    1 Error
*/
int Init(U32 Addr, U32 Freq, U32 Func) {
  U32 v;
  (void)Addr;
  (void)Freq;
  (void)Func;
  
  _pRestoreInfo.acr = FLASH->ACR & 0x00001F07;
  FLASH->ACR = 7; // Disable all caches and set wait-states to 7 to be on the safe side
  do {
    _FeedWatchdog();
    v = FLASH->ACR;
    if ((v & 0x00001F07) == 7) {  // Wait until write has completely succeeded
      break;
    }
  } while (1);
  FLASH->ACR = (1 << 11) | (1 << 12) | 7;  // Reset data and instruction cache
  FLASH->ACR = 7;                          // Invalidate reset of data and instruction cache
  return 0;
}

/*********************************************************************
*
*       UnInit
*
*  Function description
*    Handles the de-initialization of the flash module.
*
*  Parameters
*    Func: Caller type (e.g.: 1 - Erase, 2 - Program, 3 - Verify)
*
*  Return value 
*    0 O.K.
*    1 Error
*/
int UnInit(U32 Func) {
  U32 v;
  (void)Func;
  //
  // Reset data and instruction cache
  //
  FLASH->ACR = (1 << 11) | (1 << 12) | 7;  // Reset data and instruction cache
  FLASH->ACR = 7;  // Invalidate reset of data and instruction cache
  //
  // Restore cache settings
  //
  FLASH->ACR = _pRestoreInfo.acr; // Restore flash wait states and cache settings
  do { 
    _FeedWatchdog();
    v = FLASH->ACR;
    if ((v & 0x00001F07) == _pRestoreInfo.acr) {  // Wait until write has completely succeeded
      break;
    }
  } while (1);
  return 0;
}

/*********************************************************************
*
*       EraseSector
*
*  Function description
*    Erases one flash sector.
*
*  Parameters
*    Addr: Address of the sector to be erased
*
*  Return value 
*    0 O.K.
*    1 Error
*/
int EraseSector(U32 SectorAddr) {
  U32 Status;
  U32 NumSectorsRem;
  U32 SectorIndex;
  int r;
  
  _FeedWatchdog();
  r = 0;
  SectorIndex = SectorAddr >> 14; // 4KB
  //
  // Clear all error bits
  //
  FLASH->SR = 0
            | (1 << 1)   // Clear OPERR bit by writing 1 to it
            | (1 << 4)   // Clear WRPERR bit by writing 1 to it
            | (1 << 5)   // Clear PGAERR bit by writing 1 to it
            | (1 << 6)   // Clear PGPERR bit by writing 1 to it
            | (1 << 7)   // Clear PGSERR bit by writing 1 to it
            ;
  //
  // Unlock the flash control register
  //
  FLASH->KEYR = FLASH_KEY1;
  FLASH->KEYR = FLASH_KEY2;
  //
  // Setup flash controller
  //
  FLASH->CR   = (1uL << 1) | (SectorIndex << 3);    // Set flash controller to sector erase mode and select sector to be erased
  FLASH->CR  |= (1uL << 16);   // Start erase operation
  //
  // Wait for operation to complete
  //
  //  Status = *((volatile U32 *)(0x20000000));
  //  Status = *((volatile U32 *)(0x20000000));
  do {
    _FeedWatchdog();
    Status = FLASH->SR;
  } while (Status & (1uL << 16));
  FLASH->CR = (1uL << 31);        // Lock access to flash control register. Not entirely sure why this is needed, but it seems if we do not lock the flash control register again, flash programming sometimes fails..
  if (Status & 0xFE) {
    //
    // Clear all error bits
    //
    FLASH->SR = 0
              | (1 << 1)   // Clear OPERR bit by writing 1 to it
              | (1 << 4)   // Clear WRPERR bit by writing 1 to it
              | (1 << 5)   // Clear PGAERR bit by writing 1 to it
              | (1 << 6)   // Clear PGPERR bit by writing 1 to it
              | (1 << 7)   // Clear PGSERR bit by writing 1 to it
              ;
    r = 7;
  }
  return r;
}

#if 0  // We do not support EraseChip() as it is much slower than erase the flash sector by sector
/*********************************************************************
*
*       EraseChip
*
*  Function description
*    Erases the entire flash
*
*  Return value 
*    0 O.K.
*    1 Error
*/
int EraseChip(void) {
  U32 Status;
  //
  // Clear all error bits
  //
  FLASH->SR = 0
            | (1 << 1)   // Clear OPERR bit by writing 1 to it
            | (1 << 4)   // Clear WRPERR bit by writing 1 to it
            | (1 << 5)   // Clear PGAERR bit by writing 1 to it
            | (1 << 6)   // Clear PGPERR bit by writing 1 to it
            | (1 << 7)   // Clear PGSERR bit by writing 1 to it
            ;
  //
  // Unlock the flash control register
  //
  FLASH->KEYR = FLASH_KEY1;
  FLASH->KEYR = FLASH_KEY2;
  FLASH->CR |=  (1uL << 2);                              // Mass Erase Enabled
  FLASH->CR |=  (1uL << 16);                             // Start Erase
  //
  // Wait for operation to complete
  //
  do {
    _FeedWatchdog();
    Status = FLASH->SR;
  } while (Status & (1uL << 16));
  FLASH->CR &= ~(1uL << 2);                              // Mass Erase Disabled
  return 0;
}
#endif

/*********************************************************************
*
*       ProgramPage
*
*  Function description
*    Programs one flash page.
*
*  Parameters
*    DestAddr: Destination address
*    NumBytes: Number of bytes to be programmed (always a multiple of program page size, defined in FlashDev.c)
*    pSrcBuff: Point to the source buffer
*
*  Return value 
*    0 O.K.
*    1 Error
*/
int ProgramPage(U32 DestAddr, U32 NumBytes, U8 *pSrcBuff) {
  volatile U8 * pSrc;
  volatile U8 * pDest;
  U8 AccessWidth;
  U32 Status;
  int r;

  _FeedWatchdog();
  r           = 0;
  pSrc        = (volatile U8*)pSrcBuff;
  pDest       = (volatile U8*)DestAddr;
  AccessWidth = 4;    // VTarget has to be >= 2.7V
  //
  // Clear all error bits
  //
  FLASH->SR = 0
            | (1 << 1)   // Clear OPERR bit by writing 1 to it
            | (1 << 4)   // Clear WRPERR bit by writing 1 to it
            | (1 << 5)   // Clear PGAERR bit by writing 1 to it
            | (1 << 6)   // Clear PGPERR bit by writing 1 to it
            | (1 << 7)   // Clear PGSERR bit by writing 1 to it
            ;
  //
  // Unlock the flash control register
  //
  FLASH->KEYR = FLASH_KEY1;
  FLASH->KEYR = FLASH_KEY2;
  //
  // Setup flash controller
  //
  FLASH->CR = ((AccessWidth >> 1) << 8);
  FLASH->CR |= (1 << 0);
  //
  // Program items
  //
  do {
    *((U32 *)pDest) = *((U32 *)pSrc);
    //
    // Wait for operation to complete
    //
    do {
      _FeedWatchdog();
      Status = FLASH->SR;
    } while (Status & (1uL << 16));
    if (Status & 0xFE) {
      r = 4;
      goto Done;
    }
    pSrc     += AccessWidth;
    pDest    += AccessWidth;
    NumBytes -= AccessWidth;
  } while(NumBytes);
Done:
  FLASH->CR = (1uL << 31); // Lock access to flash control register. Not entirely sure why this is needed, but it seems if we do not lock the flash control register again, flash programming sometimes fails..
  //
  // Clear all error bits
  //
  FLASH->SR = 0
            | (1 << 1)   // Clear OPERR bit by writing 1 to it
            | (1 << 4)   // Clear WRPERR bit by writing 1 to it
            | (1 << 5)   // Clear PGAERR bit by writing 1 to it
            | (1 << 6)   // Clear PGPERR bit by writing 1 to it
            | (1 << 7)   // Clear PGSERR bit by writing 1 to it
            ;
  return r;
}